"""Contains opsin models, light sources, and some parameters"""
from cleo.opto.opsins import (
    FourStateOpsin,
    ProportionalCurrentOpsin,
    Opsin,
    BansalFourStateOpsin,
    # BansalThreeStatePump,
)
from cleo.opto.opsin_library import (
    chr2_4s,
    chr2_b4s,
    vfchrimson_4s,
    chrimson_4s,
    gtacr2_4s,
)
